# -*- Mode: Python; tab-width: 4 -*-
#
#	$Id: FSSession.py,v 2.6 2000/06/29 17:27:21 pavlos Exp $
#	Author: Pavlos Christoforou <pavlos@gaaros.com>
#
# ======================================================================
#
#
#  Disclaimer
#
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#     The name of the author or GAAROS may not be used to endorse or
#     promote products derived from this software without specific prior
#     written permission
#
#
#   THIS SOFTWARE IS PROVIDED BY GAAROS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL GAAROS OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
########################################################################

__version__='$Revision: 2.6 $'[11:-2]

import os,string,time
import cPickle
import whrandom
import App.FindHomes
from UserDict import UserDict

import Globals

from Shared.DC.ZRDB.TM import TM
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from App.Management import Navigation, Tabs
from Acquisition import Implicit
import OFS.SimpleItem 
from OFS.Document import Document
import AccessControl.Role 
import DateTime

# Define load/dump functions
_dump=cPickle.dump
_load=cPickle.load

FSSessionDir=os.path.join(INSTANCE_HOME,'var','FSSession')

class Store(Implicit,TM):
    """Implements a non persistent object that is transaction aware"""

    ## Transaction Control

    def tpc_begin(self,*ignored):
        # First create a temporary file with session info
        self.aq_parent.dump()
        

    def _abort(self,*ignored):
        # Maybe I should remove the temporary file here.
        pass
    
    def _finish(self,*ignored):
        self.aq_parent.final_commit()

class FSSession(OFS.SimpleItem.Item, Persistent, Implicit, 
		      AccessControl.Role.RoleManager):
    """ A File system based session management product """

    meta_type='FSSession'
    id       ='FSSession'
    title    ='FSSession'
    icon     ='misc_/FSSession/fssession'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_tabs' ]),
    )

    manage_main=HTMLFile('main', globals())

    def __init__(self, cookie_name):
        self.cookie_name = cookie_name
        
    def __call__(self,noCookie=None):
        '''Initiate session'''
        self._v_store=Store()
        self._v_store._register()
        self._v_data={}
        self._v_SessionUIDChanged=0
        self.load()
        if noCookie:
            self._v_cookie_option=0
        else:
            self._v_cookie_option=1
            
        
    def load(self,tm=time,jn=os.path.join):
        
        if ((self.REQUEST.has_key('_force_new_session') and
            self.REQUEST['_force_new_session']) or not
            self.REQUEST.has_key(self.cookie_name)): # comply with SQLSession
            # create new UID
            suffix=whrandom.randint(1,100000000)
            t=tm.gmtime(tm.time())
            self._v_uid=uid='%s%i'%(tm.strftime('%j%I%M%S',t),suffix)
            self.REQUEST[self.cookie_name]=self._v_uid
            self._v_SessionUIDChanged=1
        else:
            self._v_uid=uid=self.REQUEST[self.cookie_name]
            day=jn(FSSessionDir,uid[:3])
            hour=jn(day,uid[3:5])
            minute=jn(hour,uid[5:7])

            try:
                f=open(jn(minute,uid),'rb')
                self.update(_load(f))
                f.close()
            except IOError:
                # create new session
                self.clear()
            except:
                # zero length file, or crazy pickled object like an external
                # method. Remove bloody file and reraise exception
                f.close()
                os.remove(jn(minute,uid))
                raise
                
    def dump(self,tm=time,ex=os.path.exists,jn=os.path.join):

        day=jn(FSSessionDir,self._v_uid[:3])
        hour=jn(day,self._v_uid[3:5])
        minute=jn(hour,self._v_uid[5:7])
    
        if not ex(day):
            os.mkdir(day)
        if not ex(hour):
            os.mkdir(hour)
        if not ex(minute):
            os.mkdir(minute)

        if self._v_cookie_option and self._v_SessionUIDChanged:
            # cookie expires after one year
            t=DateTime.DateTime()+365
            t_expires='%s,%s-%s-%s 23:59:59 GMT'%(t.Day(),
                                                  t.dd(),t.aMonth(),t.yy())
            self.REQUEST.RESPONSE.setCookie(self.cookie_name,
                         self._v_uid,expires=t_expires,path='/')

        self._v_filename=jn(minute,self._v_uid)
        tmpfile=self._v_filename+'.tmp'
        f=open(tmpfile,'wb')
        try:
            _dump(self._v_data,f)
        except:
            f.close()
            os.remove(tmpfile)
            raise
        else:
            f.close()

    def final_commit(self,jn=os.path.join):
        '''This is now an atomic operation. Any possible errors like
        failure to Pickle full filesystem etc would have raised an
        exception in tpc_begin'''

        # should never fail
        os.rename(self._v_filename+'.tmp',self._v_filename)

        
    def url(self,url):
        if self.REQUEST.environ.has_key('HTTP_COOKIE'):
            # Client accepts cookies. Just return url. A situation can
            # arise where the client does accept cookies but there is
            # no cookie set for our domain. This method will mistake
            # the client as a non cookie enabled, but only the first
            # time.
            return url
        else:
            if '?' in url:
                return '%s&%s=%s'%(url,self.cookie_name,self._v_uid)
            else:
                return '%s?%s=%s'%(url,self.cookie_name,self._v_uid)

    def getName(self):
        'Return SessionUID'
        return self._v_uid

    def __setstate__(self,state):
        'Fix stupidities of old ...'
        Persistent.__setstate__(self,state)
        if hasattr(self,'uid'):
            del self.uid

    ## Use _v_data in dict methods
    ## __getattr_ would be more elegant but with ZODB one has to be
    ## careful

    def __repr__(self): return repr(self._v_data)
    def __cmp__(self, dict):
        if isinstance(dict, UserDict):
            return cmp(self._v_data, dict._v_data)
        else:
            return cmp(self._v_data, dict)
    def __len__(self): return len(self._v_data)
    def __setitem__(self, key, item): self._v_data[key] = item
    def __delitem__(self, key): del self._v_data[key]
    def clear(self): self._v_data.clear()
    def keys(self): return self._v_data.keys()
    def items(self): return self._v_data.items()
    def values(self): return self._v_data.values()
    def has_key(self, key): return self._v_data.has_key(key)
    def update(self, dict):
        if isinstance(dict, UserDict):
            self._v_data.update(dict.data)
        elif isinstance(dict, type(self._v_data)):
            self._v_data.update(dict)
        else:
            for k, v in dict.items():
                self._v_data[k] = v

    def __getitem__(self, key):
        # Do  not return error if key does not exist
        return self._v_data.get(key)

    def copy(self):
        # copy doesn't make much sense in this context
        raise AttributeError,'Copy not supported for Session storage objects'

    set=__setitem__

    delete=__delitem__

    def __str__(self):
        '''Pretty print contents'''
        out='<B>FSSession Contents:</B><BR>\n'
        for k,v in self.items():
            out='%s\t<LI>%s:\t%s\n'%(out,str(k),str(v))
        return out



manage_addFSSessionForm=HTMLFile('addFSSession', globals())

def manage_addFSSession(self, cookie_name, REQUEST=None):
    """ Code from Anthony Baxter, more or less"""
    if hasattr(self.aq_base, 'FSSession'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a FSSession',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    ob=FSSession(cookie_name)
    self._setObject('FSSession', ob)
    if REQUEST:
        return self.manage_main(self,REQUEST)


